/*
 * implementation of class Filed::Dipole
 * $Author: claudio $
 *
 * $Revision: 1.13 $
 *
 * $Log: dipole.cpp,v $
 * Revision 1.13  2018-11-15 15:23:42  claudio
 * start release_8 developement
 *
 * Revision 1.12  2009/06/16 14:07:26  claudio
 * all kicks in [rad] ( formerly [mrad] )
 *
 * Revision 1.11  2009/06/11 13:56:07  claudio
 * new formula for strength calculation, electron rest mass is considered for momentum calculation
 *
 * Revision 1.10  2009/03/31 07:57:02  claudio
 * Major change: Strength now referes to the deflection angle.
 * added methods to handle triajectory curvature - formerly strength -
 *
 * Revision 1.9  2009/02/03 16:08:55  claudio
 * corrected bug in _updateRange
 *
 * Revision 1.8  2009/01/07 08:45:02  claudio
 * naturalSpline class and files replaced by Spline - more generic
 *
 * Revision 1.7  2009/01/05 12:56:58  claudio
 * doxygen comments
 *
 * Revision 1.6  2007/08/06 13:44:44  claudio
 * added functions to exactly calculate the inverse of the ineterpolator. based on the "regula falsi" root finding algorithm and
 * a bracleting strategy
 *
 * Revision 1.5  2007/05/18 12:35:22  claudio
 * added field to handle positive bending field
 *
 * Revision 1.4  2007/03/28 15:11:45  claudio
 * refactor towards unification of common code
 *
 * Revision 1.3  2007/03/28 13:19:54  claudio
 * development
 *
 * Revision 1.2  2007/03/26 15:04:28  claudio
 * development step of end March
 *
 * Revision 1.1.1.1  2007/03/14 10:27:13  claudio
 * Hlf field module
 */

//TODO: chack that geometric length is always >0!!

#include "dipole.h"
#include "spline.h"
#include "multipolynomial.h"
#include <cmath>
#include <limits>
#include <algorithm>

namespace magnetfield
{
using namespace Interpolator;

Dipole::Dipole(bool positive, Interpolator::Interpolator& direct_interpolator,Interpolator::Interpolator& inverse_interpolator,Interpolator::Interpolator& length_interpolator,Interpolator::Interpolator& energy_interpolator, Interpolator::Interpolator& inverse_energy_interpolator, std::string name,std::string psname):_gradient(0),_fieldIndex(0),_radius(NAN),_angle(NAN)
{
	_positive=positive;
	_name=name;
	_psName=psname;
	_fieldType=f_dipole;
	
	//use dynamic_cast to build an actual copy of the effective type without breaking encapsulation.
	// On the other hand there is no transfer of ownership of objects: better insulation, no problems with automatic objetcs going out of scope
	
	_principal=0;   //set guard value
	if( Spline* spline=dynamic_cast<Spline*>(&direct_interpolator) ){
		Spline *myspline=new Spline(*spline);
		_principal=myspline;
	}
	if( multiPolynomial* poly=dynamic_cast<multiPolynomial*>(&direct_interpolator) ){
		multiPolynomial *mypoly=new multiPolynomial(*poly);
		_principal=mypoly;
	}
	if(_principal==0) throw std::invalid_argument("Dipole(): direct_interpolator");
	
	_inverse=0; //set guard value
	if( Spline* spline=dynamic_cast<Spline*>(&inverse_interpolator) ){
		Spline *myspline=new Spline(*spline);
		_inverse=myspline;
	}
	if( multiPolynomial* poly=dynamic_cast<multiPolynomial*>(&inverse_interpolator) ){
		multiPolynomial *mypoly=new multiPolynomial(*poly);
		_inverse=mypoly;
	}
	if(_inverse==0){
		delete _principal;
		throw std::invalid_argument("Dipole(): inverse_interpolator");
	}
	
	_length=0; //set guard value
	if( Spline* spline=dynamic_cast<Spline*>(&length_interpolator) ){
		Spline *myspline=new Spline(*spline);
		_length=myspline;
	}
	if( multiPolynomial* poly=dynamic_cast<multiPolynomial*>(&length_interpolator) ){
		multiPolynomial *mypoly=new multiPolynomial(*poly);
		_length=mypoly;
	}
	if(_length==0){
		delete _principal;
		delete _inverse;
		throw std::invalid_argument("Dipole(): length_interpolator");
	}
	_energy=0; //set guard value
	if( Spline* spline=dynamic_cast<Spline*>(&energy_interpolator) ){
		Spline *myspline=new Spline(*spline);
		_energy=myspline;
	}
	if( multiPolynomial* poly=dynamic_cast<multiPolynomial*>(&energy_interpolator) ){
		multiPolynomial *mypoly=new multiPolynomial(*poly);
		_energy=mypoly;
	}
	if(_energy==0){
		delete _principal;
		delete _inverse;
		delete _length;
		throw std::invalid_argument("Dipole(): energy_interpolator");
	}
	_invEnergy=0; //set guard value
	if( Spline* spline=dynamic_cast<Spline*>(&inverse_energy_interpolator) ){
		Spline *myspline=new Spline(*spline);
		_invEnergy=myspline;
	}
	if( multiPolynomial* poly=dynamic_cast<multiPolynomial*>(&inverse_energy_interpolator) ){
		multiPolynomial *mypoly=new multiPolynomial(*poly);
		_invEnergy=mypoly;
	}
	if(_invEnergy==0){
		delete _principal;
		delete _inverse;
		delete _length;
		delete _energy;
		throw std::invalid_argument("Dipole(): inverse_energy_interpolator");
	}
	_updateRange();
	
}

Dipole::~Dipole( )
{
	delete _principal;
	delete _inverse;
	delete _length;
	delete _energy;
	delete _invEnergy;
	if(_gradient) delete _gradient;
	if(_fieldIndex) delete _fieldIndex;
}
double Dipole::excitationToEnergy (double excitation )
{
	return _energy->evaluate(excitation);
}


double Dipole::energyToExcitation(double energyGeV )
{
	double exc= _invEnergy->evaluate(energyGeV);
	return _energy->inverse_evaluate(energyGeV,exc);
}


double Dipole::excitationToField (double excitation )
{
	if( _positive )
		return _principal->evaluate(excitation);
	else
		return -_principal->evaluate(excitation);
}

double Dipole::excitationToMagneticLength(double excitation )
{
	return _length->evaluate(excitation);
}

double Dipole::excitationToCurvature (double excitation, double energyGeV )
{
	if(energyGeV <= 0.0) throw std::range_error("excitationToCurvature(): negative or null energy"); //sanity check
	double field=_principal->evaluate(excitation);
	double curvature= field*(c_light_gev/p_gev(energyGeV)); //TODO: sistemare formula: serve curvatura!!
	if( _positive )
		return curvature;
	else
		return -curvature;
}



double Dipole::excitationToStrength (double excitation, double energyGeV )
{
	if(energyGeV <= 0.0) throw std::range_error("excitationToCurvature(): negative or null energy"); //sanity check
	double field=_principal->evaluate(excitation);
	double maglen=_length->evaluate(excitation);
	double deflection= (field*maglen*(c_light_gev/p_gev(energyGeV)));
	if( _positive )
		return deflection;
	else
		return -deflection;
}

double Dipole::fieldToExcitation (double field )
{
	if(_positive){
		double exc=_inverse->evaluate(field);
		return _principal->inverse_evaluate(field,exc);
	}
	else{
		double exc=_inverse->evaluate(-field);
		return _principal->inverse_evaluate(-field,exc);
	}
}


double Dipole::curvatureToExcitation (double curvature, double energyGeV )
{
	if(energyGeV <= 0.0) throw std::range_error("curvatureToExcitation(): negative or null energy"); //sanity check
	double field;
	if(_positive)
		field = curvature/(c_light_gev/p_gev(energyGeV));
	else
		field = -curvature/(c_light_gev/p_gev(energyGeV));

	double exc=_inverse->evaluate(field);
	return _principal->inverse_evaluate(field,exc);
	//return _inverse->evaluate(field);
}



double Dipole::strengthToExcitation (double strength, double energyGeV )
{
	//Attenzione! Il calcolo non tiene conto delle variazioni di lunghezza magnetica con l' eccitazione
	
	if(energyGeV <= 0.0) throw std::range_error("Dipole::strengthToExcitation(): negative or null energy"); //sanity check
	if(_geomLength <=0.0) throw std::range_error("Dipole::strengthToExcitation(): negative or null geomtric length"); //sanity check
	double field;
	double radians=strength;
	if(_positive)
		field = (radians/(c_light_gev/p_gev(energyGeV)))/_geomLength;
	else
		field = -(radians/(c_light_gev/p_gev(energyGeV)))/_geomLength;

	double exc=_inverse->evaluate(field); //get first guess excitation
	return _principal->inverse_evaluate(field,exc); //refine excitation estimate 
	//return _inverse->evaluate(field);
}


void Dipole::getExcitationRange(double& min, double& max )
{
	//_principal->get_range(min,max);
	min=_minExcitation;
	max=_maxExcitation;
	return;
}


void Dipole::getFieldRange(double& min, double& max )
{
	if(_positive){
		min=excitationToField(_minExcitation);
		max=excitationToField(_maxExcitation);
	}
	else{
		min=excitationToField(_maxExcitation);
		max=excitationToField(_minExcitation);
	}
	
	return;
}


void Dipole::getCurvatureRange(double& min, double& max, const double energyGeV )
{
	if(energyGeV <= 0.0) throw std::range_error("getCurvatureRange(): negative or null energy"); //sanity check
	if(_positive){
		min=excitationToField(_minExcitation)*(c_light_gev/p_gev(energyGeV));
		max=excitationToField(_maxExcitation)*(c_light_gev/p_gev(energyGeV));
	}
	else{
		min=excitationToField(_maxExcitation)*(c_light_gev/p_gev(energyGeV));
		max=excitationToField(_minExcitation)*(c_light_gev/p_gev(energyGeV));
	}
	return;
}


void Dipole::getStrengthRange(double& min, double& max, const double energyGeV )
{
	if(energyGeV <= 0.0) throw std::range_error("getStrengthRange(): negative or null energy"); //sanity check
	double minLen,maxLen,minField,maxField;
	if(_positive){
		minLen=_length->evaluate(_minExcitation);
		maxLen=_length->evaluate(_maxExcitation);
		minField=excitationToField(_minExcitation);
		maxField=excitationToField(_maxExcitation);
		min=(minField*minLen*(c_light_gev/p_gev(energyGeV)));
		max=(maxField*maxLen*(c_light_gev/p_gev(energyGeV)));
	}
	else{
		minLen=_length->evaluate(_maxExcitation);
		maxLen=_length->evaluate(_minExcitation);
		minField=excitationToField(_maxExcitation);
		maxField=excitationToField(_minExcitation);
		min=(minField*minLen*(c_light_gev/p_gev(energyGeV)));
		max=(maxField*maxLen*(c_light_gev/p_gev(energyGeV)));
	}
	return;
}


void Dipole::getEnergyRange(double& min, double& max)
{
	min=excitationToEnergy(_minExcitation);
	max=excitationToEnergy(_maxExcitation);
	return;
}

void Dipole::setRadius(double rho)
{
	if(rho<=0.0) throw std::range_error("Set radius(): negative or null radius of curvature"); //sanity check
	_radius=rho;
	return;
}

double Dipole::getRadius()
{
	return _radius;
}

void Dipole::setAngle(double phi)
{
	if(phi< 0.0 || phi >360.0 ) throw std::range_error("Set angle(): agle out of range"); //sanity check
	_angle=phi;
	return;
}

double Dipole::getAngle()
{
	return _angle;
}

double Dipole::getGradient(const double curr) const
{
	if (_gradient) return _gradient->evaluate(curr);
	else return 0.0;
}


double Dipole::getIndex(const double curr) const
{
	if (_fieldIndex) return _fieldIndex->evaluate(curr);
	else return 0.0;
}

void Dipole::setGradientInterpolator(Interpolator::Interpolator& gradient_interpolator)
{
	if(_gradient) delete _gradient; //check for prevoius installation
	if( Spline* spline=dynamic_cast<Spline*>(&gradient_interpolator) ){
		Spline *myspline=new Spline(*spline);
		_gradient=myspline;
	}
	if( multiPolynomial* poly=dynamic_cast<multiPolynomial*>(&gradient_interpolator) ){
		multiPolynomial *mypoly=new multiPolynomial(*poly);
		_gradient=mypoly;
	}
	_updateRange();
}

void Dipole::setIndexInterpolator(Interpolator::Interpolator& index_interpolator)
{
	if(_fieldIndex) delete _fieldIndex; //check for prevoius installation
	if( Spline* spline=dynamic_cast<Spline*>(&index_interpolator) ){
		Spline *myspline=new Spline(*spline);
		_fieldIndex=myspline;
	}
	if( multiPolynomial* poly=dynamic_cast<multiPolynomial*>(&index_interpolator) ){
		multiPolynomial *mypoly=new multiPolynomial(*poly);
		_fieldIndex=mypoly;
	}
	_updateRange();
}

//!check the various ranges and set the minimum corretc range for excitation
void Dipole::_updateRange()
{
	double dmin,dmax;
	double emin,emax;
	double gmin,gmax;
	double fimin,fimax;
	double tmin,tmax;
	_principal->get_range(dmin,dmax);
	_energy->get_range(emin,emax);
	
	if(_gradient) _gradient->get_range(gmin,gmax);
	else{
		gmin=-std::numeric_limits<double>::max();// min() is the smallest positive double which is different from 0 !!
		gmax=std::numeric_limits<double>::max();
	}
	if(_fieldIndex) _fieldIndex->get_range(fimin,fimax);
	else{
		fimin=-std::numeric_limits<double>::max(); // min() is the smallest positive double which is different from 0 !!
		fimax=std::numeric_limits<double>::max();
	}
	
	tmin=std::max(dmin,emin);
	
	tmin=std::max(tmin,gmin);
	
	
	tmin=std::max(tmin,fimin);
	
	
	tmax=std::min(dmax,emax);
	tmax=std::min(tmax,gmax);
	tmax=std::min(tmax,fimax);
	
	_minExcitation=tmin;
	_maxExcitation=tmax;
}

} //close namespace Filed


