
#ifndef FIELD_MAGNET_H
#define FIELD_MAGNET_H

#include <string>
#include <interpolator.h>

/**
* @class Magnet
* Model  of an accelerator magnet main quantitites and fields.
* Main duties are to provide:
* Static data about the magnet: geometry, position, survey errors. (to be done)
* Operational data: "name" of power supply: attribute or command name,  or rpc F_M_A_M for
* connection to operational tango or rpc device. It is assumed that there is only
* one quantity current [A,kA,mA] or voltage [V,kV,mV] that drives the magnet.
* calculated data: field determined by current/voltage, current/voltage needed to
* produce a field.
* Extra data: higher order and skew fields determined by gcurrent/voltage. 0 if
* not available. (to be done)
*/

/* $Author: claudio $
*
* $Revision: 1.18 $
*
* $Log: magnet.h,v $
* Revision 1.18  2018-11-15 15:23:42  claudio
* start release_8 developement
*
* Revision 1.17  2010/01/15 16:30:47  claudio
* FocussingSignConvention: MAD (K>0) or Vicky (K<0) for quads and sexts.
*
* Revision 1.16  2009/06/11 15:54:50  claudio
* added empirical strength correction factor
*
* Revision 1.15  2009/06/11 13:56:07  claudio
* new formula for strength calculation, electron rest mass is considered for momentum calculation
*
* Revision 1.14  2009/03/31 07:54:10  claudio
* minor doc corrections
*
* Revision 1.13  2009/01/07 08:42:57  claudio
* adapted to new class names for interpolators
*
* Revision 1.12  2009/01/05 12:59:42  claudio
* only one instance of c_light_GeV
* doxygen comments
*
* Revision 1.11  2008/08/27 13:23:43  claudio
* doxygen improvements
*
* Revision 1.10  2007/07/24 13:09:52  claudio
* cleanup of #ifdef guard against double inclusion
* even across namespaces
*
* Revision 1.9  2007/07/23 08:18:06  claudio
* exposed "focussing" class variable via  bool getFocussing()
*
* Revision 1.8  2007/07/13 10:46:47  claudio
* modified comments/docs
*
* Revision 1.7  2007/07/09 09:13:53  claudio
* Change som #includes for compatibility with canonical installation tree
*
* Revision 1.6  2007/05/18 12:36:41  claudio
* refactored for sharing more internal variables and, later, code
*
* Revision 1.5  2007/03/30 12:55:51  claudio
* Costrusice decentemente le classi base. Migliorata la documantazione. TODO: gestire meglio i range.
* Mettere a fattore comune (rifattorizzare) il piu' possibile in magnet
*
* Revision 1.4  2007/03/28 14:55:37  claudio
* refactor towards unification of common code
*
* Revision 1.3  2007/03/28 13:19:55  claudio
* development
*
* Revision 1.2  2007/03/26 15:04:28  claudio
* development step of end March
*
* Revision 1.1.1.1  2007/03/14 10:27:13  claudio
* Hlf field module
*/




namespace magnetfield
{

enum magneticFieldType{f_dipole,f_quadrupole,f_sextupole,f_integdipole,f_solenoid};
enum unitOfMeasure{m_dipole, m_quadrupole, m_sextupole, m_integdipole, m_energy, m_dipoleindex, m_length, m_a_u};

/**
 * Two sign conventions for focussing quadrupoles an sextupoles (horizontal plane)
 * foc_sign_vicky: K <0
 * foc_sign_mad  : K >0
 */

enum focSignConvention{ foc_sign_vicky,foc_sign_mad };

static const double c_light_gev=0.299792458;  /**< conversion factor speed_of_light with energy in GeV */
static const double E_rest_gev=0.00051099891; /**< elctron rest mass in GeV */
static const double E_rest_gev_squared=2.611198860211881e-07; /**< elctron rest mass in GeV squared */


/**
 * Fcousing sign convention is global to the namespace. It must be set before the creation of 
 * of any object derived fromFiled::Magnet, either diretcly or by means of factory objects.
 * The default is to have K < 0 for horizontally focussing quadrupoles and sextupoles.
 * Be aware that this selector is shared and absolutely non thread-safe! It's inteded to be set
 * only once at application startup.
 * The problem arises when using different sign conventions inside a single instance of a tango device server; in thsi case
 * the startup order may (will ...) change the sign convention in unespected ways.
 */





/**
 * calculate the electron longitudinal momentum given the total electron energy
 * @param total electron energy [GeV]
 * @return electron momentum [GeV/c]
 */
double p_gev(double en);


/**
  * @class surveyError
  * Holder of the alignment error of an element. All members are public (like a struct)
  * N.B.: we should have a clear and uniform coordinate system and sign convention
  * established
  */

class surveyError
{
public:
	/**
	 * @param misal_x horizontal alignment error
	 */

	double misal_x ;
	/**
	 * @param rot_x error angle around x
	 */
	double rot_x;

	/**
	 * @param misal_y vertical alignment error
	 */
	double misal_y;

	/**
	 * @param rot_x error angle around x
	 */
	double rot_y;

	/**
	 * @param misal_s longitudinal alignment error
	 */
	double misal_s;

	/**
	 * @param rot_s error angle around s
	 */
	double rot_s;

	inline surveyError():misal_x(0.0),rot_x(0.0),misal_y(0.0),rot_y(0.0),misal_s(0.0),rot_s(0.0){};
};

class Magnet
{
public:

	Magnet( );

	/**
	 * Empty virtual Destructor
	 */
	virtual ~Magnet( );

	/**
	 * @return name of magnet
	 */
	std::string getName( ) const;


	/**
	 * @return name of power supply driving the magnet
	 */
	std::string getPsName( ) const;


	/**
	 * @return surveyError of the magnet
	 */
	surveyError getSurveyError( ) const;

	/**
	 * @return the geometric Length of the magnet
	 */
	double getGeometricLength( ) const;

	/**
	 * Set the geometric lenght of the magnet
	 * @param newLength the new value of _geomLength
	 */
	void setGeometricLength(const double newLength);
	
	/**
	 * Get focussing/defocussing property. Apllies mainly to quadrupoles an sextupoles
	 * focussing means gradient , strenght,etc.. are < 0 for poisitive excitation:
	 * - (focussing ==  true && I > 0) ==> k <0
	 * - (focussing == false && I < 0) ==> k >0
	 * By default the magnet is focussing
	 */
	
	bool isFocussing() const;
	
	/**
	 * Get positivie/negative deflection property. Applies to dipoles, kickers,correctors
	 * positive means that for I>0 electrons are deflected to the right (horizontal) or up (vertical)
	 */
	
	bool isPositive() const;


	/**
	 * Converts the excitation to the main field component of the magnet
	 * @return the main magnteic field (dipole,gradient,B*l, depending on magnet type)
	 * @param  excitation the magnet excitation ( current or  voltage , depending on magnet type )
	 * @throw std::range_error if the excitation is outside the valid range
	 */
	virtual double excitationToField(const double excitation ) =0;


	/**
	 * Converts the excitation to the magnetic length of the magnet
	 * @return the magnetic length ( 0 for B*l dmagntes like kickers)
	 * @param  excitation the excitation - current or voltage - driving the magnet. Can  be 0 for some kind of magnet
	 * @throw std::range_error if the excitation is outside the valid range
	 */
	virtual double excitationToMagneticLength(const double excitation ) =0;


	/**
	 * Converts the excitation to the strength of the magnet
	 * @return the main magnet strength (k, k', kick in milliradians depending on magnet type)
	 * @param  excitation excitation - current or voltage - driving the magnet
	 * @param  energyGeV the energy in GeV of the electron beam
	 * @throw std::range_error if the excitation is outside the valid range
	 */
	virtual double excitationToStrength(const double excitation, const double energyGeV ) =0;


	/**
	 * Converts the magnet field to the required excitation
	 * @return the required excitation ( current or  voltage , depending on magnet type )
	 * @param  field requested main magnetic field (dipole, gradient, ..., B*l, depending on magnet type)
	 * @throw std::range_error if the field is outside the valid range
	 */
	virtual double fieldToExcitation(const double field ) =0;


	/**
	 * Converts the magnet strength for a given electron energy to the required excitation
	 * @return the required excitation ( current or  voltage, depending on magnet type )
	 * @param  strength requested magnet strength (k, k', kick in milliradians, ..., depending on magnet type)
	 * @param  energyGeV the energy in GeV of the electron beam
	 * @throw std::range_error if the strenght is outside the valid range
	 */
	virtual double strengthToExcitation(const double strength, const double energyGeV ) = 0;


	/**
	 * Use for decting specific informations.
	 * @return the magnetic field type
	 */
	magneticFieldType getFieldType( ) const;


	/**
	 * Get supported range of excitation values.
	 * Range is  [min,max).
	 * @param  min ( current or  voltage , depending on magnet type )
	 * @param  max ( current or  voltage , depending on magnet type )
	 */
	virtual void getExcitationRange(double& min, double& max ) = 0;

	/**
	 * Get supported range of main field values.
	 * Range is  [min,max).
	 * @param  min (dipole,gradient, ..., B*l, depending on magnet type)
	 * @param  max (dipole,gradient, ..., B*l, depending on magnet type)
	 */
	virtual void getFieldRange(double& min, double& max ) = 0;


	/**
	 * Get supported range of field strength values for a given electron energy.
	 * Range is  [min,max).
	 * @param  min (k, k', kick in milliradians, ..., depending on magnet type)
	 * @param  max (k, k', kick in milliradians, ..., depending on magnet type)
	 * @param  energy the energy in GeV of the electron beam
	 * @throw std::range_error negative or null energy specified
	 */
	virtual void getStrengthRange(double& min, double& max, const double energy ) = 0;
	
	/**
	 * get value of empirical parameter for field compensation
	 * should help to deal with interfrence from nearby magnets and other magnetic materials
	 * suggestd by D. Castronovo
	 * Used only for strength calculations.
	 * By default it is set to 1.0
	 * @return empirical parameter
	 */
	
	double getAlpha() const;
	
	
	/**
	 * set value of the empirical parameter for field compensation
	 * @param al empirical parameter
	 */
	
	void setAlpha(const double al);

	/**
	 * read the actual global sign convention
	 */

	focSignConvention getFocSignConvetion();


	/**
	 * set the global sign convetion - to extended in quadrupoles and sextupoles
	 */

	virtual void setFocSignConvention(focSignConvention);


protected:
	bool _focussing; /**< true if magnet is focussing in the horizontal plane */
	bool _positive;  /**< true if magnet bends electrons in the positive direction */
	Interpolator::Interpolator* _principal; /**< Interpolator for the main magnetic field: f=F(i) */
	Interpolator::Interpolator* _inverse;   /**< Interpolator for the inverse of the field: i=I(f) */
	Interpolator::Interpolator* _length; /**< Interpolator for the magneti length: l=L(i). If _length is NULL, the geometric length is used */
	double _geomLength;    /**< geometric length of the magnet */
	std::string _name;     /**< name of the magnet */
	surveyError _surveyError; /**< survey error of the magnet*/
	std::string _psName;      /**< name of the power supply powering the magnet */
	double _minExcitation; /**< minimum valid excitation for the magnet */
	double _maxExcitation; /**< maximum valid excitation for the magnet */
	focSignConvention _foc_sign_convention=foc_sign_vicky; /**< sign convention property */


	// main field type
	magneticFieldType _fieldType; /**< type of the main magnetic field, i.e. class of the magnet */

	/**
	 * Set the value of _name
	 * @param newName the new value of _name
	 */
	void setName( std::string newName );

	/**
	 * Set the value of _survey_error
	 * @param newSurveyError the new value of _survey_error
	 */
	void setSurveyError( surveyError newSurveyError );

	/**
	 * Set the value of _ps_name
	 * @param newPsName the new value of _ps_name
	 */
	void setPsName( std::string newPsName );


	/**
	 * Set the value of _field_type
	 * main field type
	 * @param newFieldType the new value of _field_type
	 */
	void setFieldType( magneticFieldType newFieldType );

	/**
	 * Update and calculate the range limits
	 * to called after initialization and modifications of tables or Interpolators
	 * can be specialized in special cases
	 */

	virtual void _updateRange();
	
	/**
	 * Empirical parameter for field compensation
	 * should help to deal with interfrence from nearby magnets and other magnetic materials
	 * suggestd by D. Castronovo
	 * Used only for strength calculations for QUADRUPOLES and KICKERS: Strength_effective = Strength_table * _alpha;
	 * By default it is set to 1.0 and should not be 0.0
	 * 
	 */
	double _alpha;



};

} //close namespace Field
#endif // FIELD_MAGNET_H

