
#ifndef FIELD_MAGNETFACTORY_H
#define FIELD_MAGNETFACTORY_H

/**
* @class Field::Magnetfactory
* Factory class for automatically building a Filed::magnet object.
* In short, given a magnte name, queries the magnet database , builds and fills all the
* data needed to correctly initialize a Field::magnet, build the instance and return a pointer to the new object.
* It is the caller duty to free the Field::magnet when no longer needed (transfer of ownership)
* this class re-uses and shares a database connection
*/
 
/* $Author: claudio $
*
* $Revision: 1.12 $
*
* $Log: magnetfactory.h,v $
* Revision 1.12  2018-11-15 15:23:42  claudio
* start release_8 developement
*
* Revision 1.11  2009/03/24 09:08:39  claudio
* create solenoid
*
* Revision 1.10  2009/01/05 12:59:15  claudio
* doxygen comments
*
* Revision 1.9  2008/08/27 13:23:43  claudio
* doxygen improvements
*
* Revision 1.8  2007/07/24 13:09:52  claudio
* cleanup of #ifdef guard against double inclusion
* even across namespaces
*
* Revision 1.7  2007/07/16 08:21:18  claudio
* added support of MAGNETDB environemnt variable to pass
* usr, password and host for db connection.
*
* Revision 1.6  2007/07/09 09:13:53  claudio
* Change som #includes for compatibility with canonical installation tree
*
* Revision 1.5  2007/03/30 12:55:51  claudio
* Costrusice decentemente le classi base. Migliorata la documantazione. TODO: gestire meglio i range.
* Mettere a fattore comune (rifattorizzare) il piu' possibile in magnet
*
* Revision 1.4  2007/03/29 15:54:29  claudio
* added creation-handling of kickers
*
* Revision 1.3  2007/03/29 12:56:34  claudio
* added creation-handling of sextupoles
*
* Revision 1.2  2007/03/26 15:04:28  claudio
* development step of end March
*
* Revision 1.1.1.1  2007/03/14 10:27:13  claudio
* Hlf field module
*
*/
#include "kicker.h"
#include "sextupole.h"
#include "quadrupole.h"
#include "dipole.h"
#include "solenoid.h"
#include <stdexcept>
#include <string>

#include <mysql.h>

namespace magnetfield{

class Magnetfactory{
public:

	/**
	 * Constructor
	 * Reads the MAGNETDB environment variable to extract user,password and host for database connection
	 * MAGNETDB must be in the form     dbname:user:password:host
	 * @throw std::invalid_argument login to database failed
	 */
	Magnetfactory();

	/**
	 * Constructor
	 * @param  dbname  name od mysql magnets db (fermi_magnets, elettra_magnets,...)
	 * @param  dbuser  login name for the mysql magnets db
	 * @param  dbpasswd passworfor the mysql magnets db
	 * @param  dbhost  host name of the mysql server
	 * @throw std::invalid_argument login to database failed
	 */
	Magnetfactory(const char* dbname, const char* dbuser, const char* dbpasswd, const char* dbhost);
	~Magnetfactory();

	/**
	 * Creates and initializes a Field::Quadrupole object
	 * @param  quadname name of the wanted quadrupole
	 * @return pointer to newly allocated Quadrupole
	 * @throw std::invalid_argument invalid quadrupole name
	 * @throw std::range_error query on database returnd a wrong number of fields - check db schmema
	 */
	Quadrupole* newQuadrupole(const char* quadname);

	/**
	 * Creates and initializes a Field::Dipole object
	 * @param  quadname name of the wanted dipole
	 * @return pointer to newly allocated Dipole
	 * @throw std::invalid_argument invalid dipole name
	 * @throw std::range_error query on database returnd a wrong number of fields - check db schmema
	 */
	Dipole* newDipole(const char* quadname);


	/**
	 * Creates and initializes a Field::Sextupole object
	 * @param  sextname of the wanted sextupole
	 * @return pointer to newly allocated Sextupole
	 * @throw std::invalid_argument invalid solenoid name
	 * @throw std::range_error query on database returnd a wrong number of fields - check db schmema
	 */
	Sextupole* newSextupole(const char* sextname);
	
	
	/**
	 * Creates and initializes a Field::Solenoid object
	 * @param  solname of the wanted solenoid
	 * @return pointer to newly allocated Solenoid
	 * @throw std::invalid_argument invalid solenoid name
	 * @throw std::range_error query on database returnd a wrong number of fields - check db schmema
	 */
	Solenoid* newSolenoid(const char* solname);


	/**
	 * Creates and initializes a Field::Kicker object
	 * @param  kickname name of the wanted kicker
	 * @return pointer to newly allocated Kicker
	 * @throw std::invalid_argument invalid quadrupole name
	 * @throw std::range_error query on database returnd a wrong number of fields - check db schmema
	 */
	Kicker* newKicker(const char* kickname);

private:
	MYSQL mysql;

	/**
	 * creates the connection to the mysql database
	 * @param  dbuser  login name for the mysql magnets db
	 * @param  dbpasswd passworfor the mysql magnets db
	 * @param  dbhost  host name of the mysql server
	 * @throw std::invalid_argument login to database failed
	 */
	void initdb(const char* dbname, const char* dbuser, const char* dbpasswd, const char* dbhost);

};

} //close namespace Field


#endif //FIELD_MAGNETFACTORY_H
