/*
 * implementation of class Filed::Solenoid
 * $Author: claudio $
 *
 * $Revision: 1.4 $
 *
 * $Log: solenoid.cpp,v $
 * Revision 1.4  2018-11-15 15:23:42  claudio
 * start release_8 developement
 *
 * Revision 1.3  2010/11/11 09:43:06  claudio
 * *** empty log message ***
 *
 * Revision 1.2  2009/06/11 13:56:07  claudio
 * new formula for strength calculation, electron rest mass is considered for momentum calculation
 *
 * Revision 1.1  2009/03/24 09:09:38  claudio
 * first import, to be completed
 *
 * Revision 1.6  2009/01/07 08:42:57  claudio
 * adapted to new class names for interpolators
 *
 * Revision 1.5  2009/01/05 13:02:15  claudio
 * only one instance of c_light_GeV
 *
 * Revision 1.4  2007/08/06 13:44:44  claudio
 * added functions to exactly calculate the inverse of the ineterpolator. based on the "regula falsi" root finding algorithm and
 * a bracleting strategy
 *
 * Revision 1.3  2007/07/10 07:23:55  claudio
 * corrected sign for focusing/defocusing
 * corrected wrong SQL query
 *
 * Revision 1.2  2007/05/18 12:36:41  claudio
 * refactored for sharing more internal variables and, later, code
 *
 * Revision 1.1  2007/03/29 12:56:22  claudio
 * added creation-handling of solenoids
 *
 */

#include "solenoid.h"
#include <spline.h>
#include <multipolynomial.h>
#include <cmath>

//TODO: ps_name in database, geometric length


namespace magnetfield {
using namespace Interpolator;

Solenoid::Solenoid(bool focus,Interpolator::Interpolator& direct_interpolator,Interpolator::Interpolator& inverse_interpolator,Interpolator::Interpolator& length_interpolator,std::string name,std::string psname){
	_focussing=focus;
	_name=name;
	_psName=psname;
	_fieldType=f_solenoid;
	
	//use dynamic_cast to build an actual copy of the effective type without breaking encapsulation.
	// On the other hand there is no transfer of ownership of objects: better insulation, no problems with automatic objetcs going out of scope
	
	_principal=nullptr;   //set guard value
	if( Spline* spline=dynamic_cast<Spline*>(&direct_interpolator) ){
		Spline *myspline=new Spline(*spline);
		_principal=myspline;
	}
	if( multiPolynomial* poly=dynamic_cast<multiPolynomial*>(&direct_interpolator) ){
		multiPolynomial *mypoly=new multiPolynomial(*poly);
		_principal=mypoly;
	}
	if(_principal==nullptr) throw std::invalid_argument("Solenoid(): direct_interpolator");
	
	_inverse=nullptr; //set guard value
	if( Spline* spline=dynamic_cast<Spline*>(&inverse_interpolator) ){
		Spline *myspline=new Spline(*spline);
		_inverse=myspline;
	}
	if( multiPolynomial* poly=dynamic_cast<multiPolynomial*>(&inverse_interpolator) ){
		multiPolynomial *mypoly=new multiPolynomial(*poly);
		_inverse=mypoly;
	}
	if(_inverse==nullptr){
		delete _principal;
		throw std::invalid_argument("Solenoid(): inverse_interpolator");
	}
	
	_length=nullptr; //set guard value
	if( Spline* spline=dynamic_cast<Spline*>(&length_interpolator) ){
		Spline *myspline=new Spline(*spline);
		_length=myspline;
	}
	if( multiPolynomial* poly=dynamic_cast<multiPolynomial*>(&length_interpolator) ){
		multiPolynomial *mypoly=new multiPolynomial(*poly);
		_length=mypoly;
	}
	if(_length==nullptr){
		delete _principal;
		delete _inverse;
		throw std::invalid_argument("Solenoid(): length_interpolator");
	}
	_updateRange();
}

Solenoid::~Solenoid( ){
	delete _principal;
	delete _inverse;
	delete _length;
}


double Solenoid::excitationToField (const double excitation )
{
	if( _focussing )
		return  _principal->evaluate(excitation);
	else
		return -_principal->evaluate(excitation);
}

double Solenoid::excitationToMagneticLength(const double excitation )
{
	return _length->evaluate(excitation);
}

double Solenoid::excitationToStrength (const double excitation, const double energyGeV )
{
	if(energyGeV <= 0.0) throw std::range_error("strength(): negative or null energy"); //sanity check
	double gradient=_principal->evaluate(excitation);
	double strength= gradient*(c_light_gev/(2.0*p_gev(energyGeV)));
	if( _focussing )
		return strength;
	else
		return -strength;
}

double Solenoid::fieldToExcitation (const double field )
{
	if(_focussing){
		double exc=_inverse->evaluate(field);
		//return _inverse->evaluate(-field);
		return _principal->inverse_evaluate(field,exc);
	}
	
	else{
		//return _inverse->evaluate(field);
		double exc=_inverse->evaluate(-field);
		//return _inverse->evaluate(-field);
		return _principal->inverse_evaluate(-field,exc);
	}
}

double Solenoid::strengthToExcitation (const double strength, const double energyGeV )
{
	if(energyGeV <= 0.0) throw std::range_error("strengthToExcitation(): negative or null energy"); //sanity check
	double field;
	if(_focussing)
		field =  strength/(c_light_gev/(2.0*p_gev(energyGeV)));
	else
		field = -strength/(c_light_gev/(2.0*p_gev(energyGeV)));
	
	//return _inverse->evaluate(field);
	double exc= _inverse->evaluate(field);
	return _principal->inverse_evaluate(field,exc);
}

void Solenoid::getExcitationRange(double& min, double& max )
{
	_principal->get_range(min,max);
	return;
}

void Solenoid::getFieldRange(double& min, double& max )
{
	if(_focussing){
		_inverse->get_range(min,max);
	}
	else{
		double m,M;
		_inverse->get_range(m,M);
		min = -M;
		max = -m;
		_inverse->get_range(min,max);
		
	}
	return;
}

void Solenoid::getStrengthRange(double& min, double& max, const double energyGeV )
{
	if(energyGeV <= 0.0) throw std::range_error("getStrengthRange(): negative or null energy"); //sanity check
	double field_min,field_max;
	_inverse->get_range(field_min,field_max);
	if(_focussing){
		min=field_min*(c_light_gev/p_gev(energyGeV));
		max=field_max*(c_light_gev/p_gev(energyGeV));
	}
	else{
		min = -field_max*(c_light_gev/p_gev(energyGeV));
		max = -field_min*(c_light_gev/p_gev(energyGeV));
	}
	return;
}

} //close namespace Field
