
#ifndef FIELD_DIPOLE_H
#define FIELD_DIPOLE_H

//TODO: add index_n support for dipole 
//TODO: gestire meglio i range.

#include "magnet.h"
#include <string>

/**
  * @class Field::Dipole
  * Evaluates the main parameters of a dipole magnet: field,curvature (1/rho),nominal energy and magnetic length.
  * or estimates the excitation required for a given field or nominal energy.
  * Nominal energy is the energy of the electron which will follow the deisgn trajectory in the magnet,
  * Nominal energy is the energy of the electron which will be deflected by the design angle of the magnet.
  * In oredr to handle also combined funcitions dipoles often use in synchrtorn radiation lattices, it also estimates the gradient and field index "n" [ref].
  * By convention the field index is not calculated, but always estimated from the tables or polynomials stored in the magnets database.
  * Curvature for dipole is the curvature 1/rho for electrons.
  * These parameters are ment to be used to calculate the transfer/tranport matrices of the magnet.
  * This class exploits various Interpolator objects to evaluate the different fields, which must be
  * constructed and inited correctly. A factory class will do this reading the data from the magnets database.
  *
  * Notice that by convention all the calibration curves hold positive values for positive excitations.
  * The effective sign of the field is controlled by a boolean parameter. The value of this parametr is also recorded
  * in the database and is handled by the factory class. positve==true -> electrons are bent right on the hor. plane
  * The name of the magnet and of the power supply are auxilary fields.
  */
  
 /* $Author: claudio $
  *
  * $Revision: 1.17 $
  *
  * $Log: dipole.h,v $
  * Revision 1.17  2018-11-15 15:23:42  claudio
  * start release_8 developement
  *
  * Revision 1.16  2009/06/16 14:07:26  claudio
  * all kicks in [rad] ( formerly [mrad] )
  *
  * Revision 1.15  2009/03/31 07:57:02  claudio
  * Major change: Strength now referes to the deflection angle.
  * added methods to handle triajectory curvature - formerly strength -
  *
  * Revision 1.14  2009/01/07 08:45:02  claudio
  * naturalSpline class and files replaced by Spline - more generic
  *
  * Revision 1.13  2009/01/05 12:56:58  claudio
  * doxygen comments
  *
  * Revision 1.12  2008/08/27 13:23:43  claudio
  * doxygen improvements
  *
  * Revision 1.11  2007/08/07 15:38:53  claudio
  * aggiornate note
  *
  * Revision 1.10  2007/08/06 13:44:44  claudio
  * added functions to exactly calculate the inverse of the ineterpolator. based on the "regula falsi" root finding algorithm and
  * a bracleting strategy
  *
  * Revision 1.9  2007/07/24 13:09:52  claudio
  * cleanup of #ifdef guard against double inclusion
  * even across namespaces
  *
  * Revision 1.8  2007/07/23 08:18:59  claudio
  * exposed "_positive" class variable
  *
  * Revision 1.7  2007/07/09 09:13:53  claudio
  * Change som #includes for compatibility with canonical installation tree
  *
  * Revision 1.6  2007/05/18 12:35:22  claudio
  * added field to handle positive bending field
  *
  * Revision 1.5  2007/03/30 12:55:51  claudio
  * Costrusice decentemente le classi base. Migliorata la documantazione
  * Mettere a fattore comune (rifattorizzare) il piu' possibile in magnet
  *
  * Revision 1.4  2007/03/28 15:11:45  claudio
  * refactor towards unification of common code
  *
  * Revision 1.3  2007/03/28 13:19:54  claudio
  * development
  *
  * Revision 1.2  2007/03/26 15:04:28  claudio
  * development step of end March
  *
  * Revision 1.1.1.1  2007/03/14 10:27:13  claudio
  * Hlf field module
  *
  */



namespace magnetfield{

class Dipole : public Magnet
{
public:

	/**
	 * Constructor
	 * @param  positive select positive/negative deflection
	 * @param  direct_interpolator Interpolator object for the dipole: b=B(i)
	 * @param  inverse_interpolator Interpolator object for the current: i=I(g)
	 * @param  length_interpolator Interpolator object for the magnetic_lenght: l=L(i)
	 * @param  energy_interpolator Interpolator object for the nominal energy: E=e(i)
	 * @param  inverse_energy_intepolator Interpolator object for the current  I=I_E(e)
	 * @param  name name of the dipole magnet
	 * @param  psname name of the power supply feeding the magnet
	 * @throw std::invalid_argument Interpolator object type not detected correctly
	 */
	Dipole (bool positive, Interpolator::Interpolator& direct_interpolator,Interpolator::Interpolator& inverse_interpolator,Interpolator::Interpolator& length_interpolator,Interpolator::Interpolator& energy_interpolator, Interpolator::Interpolator& inverse_energy_intepolator,std::string name="null",std::string psname="null");

	/**
	 * Destrucor
	 */
	~Dipole ( );

	/**
	 * @return nominal energy for electrons
	 * @param  excitation magnet excitation ( current )
	 * @throw std::range_error inherited from Field::Interpolator
	 */
	double excitationToEnergy (double excitation );

	/**
	 * @return magnet excitation ( current )
	 * @param  energyGeV requested nominal energy for electrons
	 * @throw std::range_error inherited from Field::Interpolator
	 */
	double energyToExcitation (double energyGeV );

	/**
	 * @return field ( induction )
	 * @param  excitation magnet excitation ( current )
	 * @throw std::range_error inherited from Field::Interpolator
	 */
	double excitationToField (double excitation );


	/**
	 * @return magnetic length
	 * @param  excitation excitation magnet excitation ( current )
	 * @throw std::range_error inherited from Field::Interpolator
	 */
	double excitationToMagneticLength (double excitation );
	
	
	/**
	 * @return the magnet deflection (strength) [rad]
	 * @param  excitation excitation - current or voltage - driving the magnet
	 * @param  energyGeV the energy in GeV of the electron beam
	 * @throw std::range_error inherited from Field::Interpolator
	 */
	double excitationToStrength(const double excitation, const double energyGeV );




	/**
	 * @return the required excitation ( current )
	 * @param  strength requested magnet deflection (strength) [rad]
	 * @param  energyGeV the energy in GeV of the electron beam
	 * @throw std::range_error inherited from Field::Interpolator
	 */
	double strengthToExcitation(const double strength, const double energyGeV );


	/**
	 * Get supported range of magnet deflection values at a given energy.
	 * Range is  [min,max).
	 * @param  min [ rad ]
	 * @param  max [ rad ]
	 * @param  energy the energy in GeV of the electron beam
	 */

	void getStrengthRange(double& min, double& max, const double energy );
	

	/**
	 * @return curvature (1/rho) for electrons in [1/m]
	 * @param  excitation excitation magnet excitation ( current )
	 * @param  energyGeV the energy in GeV of the electron beam
	 * @throw std::range_error inherited from Field::Interpolator
	 */
	double excitationToCurvature (double excitation, double energyGeV );


	/**
	 * @return magnet excitation ( current )
	 * @param  field requested main field
	 * @throw std::range_error inherited from Field::Interpolator
	 */
	double fieldToExcitation (double field );


	/**
	 * @return excitation ( current )
	 * @param  curvature the requested curvature (1/rho) for electrons in [1/m]
	 * @param  energyGeV the energy in GeV of the electron beam
	 */
	double curvatureToExcitation (double curvature, double energyGeV )  ;


	/**
	 * returns the supported range of excitation values.
	 * Range is  [min,max].
	 * @param  min
	 * @param  max
	 */
	void getExcitationRange(double& min, double& max );

	/**
	 * returns the range of field values.
	 * Range is  [min,max].
	 * @param  min
	 * @param  max
	 */
	void getFieldRange(double& min, double& max );


	/**
	 * returns the range of curvature values (1/rho) for electrons at a given energy.
	 * Range is  [min,max].
	 * @param  min
	 * @param  max
	 * @param energy the energy in GeV of the electron beam
	 */
	void getCurvatureRange(double& min, double& max, const double energy );

	/**
	 * returns the range of valid nominal energy for the magnet
	 * Range is  [min,max].
	 * @param  min
	 * @param  max
	 */
	void getEnergyRange(double& min, double& max);

	/**
	 * set the nominal curvature radius
	 * @param rho radius of curvature [m]
	 * @throw std::range_error if radius is not of correct range
	 */
	void setRadius(double rho);

	/**
	 * get the nominal radius
	 * @return radius [m]. If not correctly set, return NaN
	 */

	double getRadius();

	/**
	 * set the nominal curvature angle
	 * @param phi angle of curvature [deg]
	 * @throw std::range_error if radius is not of correct range [0-360]
	 */
	void setAngle(double phi);

	/**
	 * get the nominal curvature angle
	 * @return phi [deg]. If not correctly set, return NaN
	 */

	double getAngle();

	/**
	 * get the gradient (qudrupolar component) of the magnet
	 * for pure bending magnets will return always 0
	 * @param curr excitation current
	 * @return the actual gardient
	 */

	double getGradient(const double curr) const;

	/**
	 * get the field index "n" of the magnet
	 * for pure bending magnets will return always 0
	 * @param curr excitation current
	 * @return the actual index
	 */
	double getIndex(const double curr) const;

	/**
	 * auxiliary method to add Interpolator for gradient
	 * to be used for combined functions dipoles. 
	 *  @param  gradient_Interpolator Interpolator object for the gradient: g=G(i)
	 */
	 //TODO make protectd but friend of factory
	void setGradientInterpolator(Interpolator::Interpolator& gradient_interpolator);

	/**
	 * auxiliary method to add Interpolator for field index
	 *  @param  index_interpolator Interpolator object for the field index: n=N(i)
	 */
	 //TODO make protectd but friend of factory
	void setIndexInterpolator(Interpolator::Interpolator& index_interpolator);


protected:
	Interpolator::Interpolator* _energy;		/**< pointer to Interpolator object for the nominal energy from current: E=e(i) */
	Interpolator::Interpolator* _invEnergy;	/**< pointer to Interpolator object for the current  from nominal energy I=I_E(e) */
	Interpolator::Interpolator* _gradient;	/**< pointer to Interpolator object for the gradient: g=G(i). Used by combined functions dipole. */
	Interpolator::Interpolator* _fieldIndex;	/**< pointer to interpolarot object for field index "n" of the magnet. Used bu combined functions dipole */
	double _radius;			/**< nominal radius of curvature */
	double _angle;			/**< nominal bending angle */
	void _updateRange();


private:


};

} //close namespace Field;
#endif // FIELD_DIPOLE_H
