#ifndef FIELD_KICKER_H
#define FIELD_KICKER_H
#include "magnet.h"
#include <string>

/**
 * @class Field::Kicker
 * Evaluates the main parameters of a kicker magnet: integrated field [T*M],strength [rad] and magnetic length
 * or estimates the excitation required for a given integrated field or strength.
 * Strength calculations are carried out for a given electron energy in GeV.
 * These parameters are ment to be used to calculate the transfer/tranport matrices of the magnet.
 * This class exploits various Interpolator objects to evaluate the different fields, which must be
 * constructed and inited correctly. A factory class will do this reading the data from the magnets database.
 *
 * Notice that by convention all the calibration curves hold POSITIVE values for POSITIVE excitations.
 * The effective sign of the field is controlled by a boolean parameter. The value of this parametr is also recorded
 * in the database and is handled by the factory class.
 * The name of the magnet and of the power supply are auxilary fields.
 */
 
/* $Author: claudio $
 *
 * $Revision: 1.13 $
 *
 * $Log: kicker.h,v $
 * Revision 1.13  2018-11-15 15:23:42  claudio
 * start release_8 developement
 *
 * Revision 1.12  2009/06/16 14:07:26  claudio
 * all kicks in [rad] ( formerly [mrad] )
 *
 * Revision 1.11  2009/03/31 07:54:59  claudio
 * minor doc corrections
 *
 * Revision 1.10  2009/01/07 08:42:57  claudio
 * adapted to new class names for interpolators
 *
 * Revision 1.9  2008/08/27 13:23:43  claudio
 * doxygen improvements
 *
 * Revision 1.8  2007/08/06 13:44:44  claudio
 * added functions to exactly calculate the inverse of the ineterpolator. based on the "regula falsi" root finding algorithm and
 * a bracleting strategy
 *
 * Revision 1.7  2007/07/24 13:09:52  claudio
 * cleanup of #ifdef guard against double inclusion
 * even across namespaces
 *
 * Revision 1.6  2007/07/23 08:19:52  claudio
 *  "_positive" class variable belongs to parent
 *
 * Revision 1.5  2007/07/13 10:46:28  claudio
 * modified comments/docs
 *
 * Revision 1.4  2007/07/09 09:13:53  claudio
 * Change som #includes for compatibility with canonical installation tree
 *
 * Revision 1.3  2007/05/18 12:36:15  claudio
 * refactored for sharing more internal variables and, later, code
 *
 * Revision 1.2  2007/03/30 12:55:51  claudio
 * Costrusice decentemente le classi base. Migliorata la documantazione. TODO: gestire meglio i range.
 * Mettere a fattore comune (rifattorizzare) il piu' possibile in magnet
 *
 * Revision 1.1  2007/03/29 15:54:29  claudio
 * added creation-handling of kickers
 *
 *
 */


namespace magnetfield
{

class Kicker : public Magnet
{
public:

	/**
	 * Constructor
	 * @param  positive select positive/negative field
	 * @param  direct_interpolator Interpolator object for the gradient: g=G(i)
	 * @param  inverse_interpolator Interpolator object for the current: i=I(g)
	 * @param  length_interpolator Interpolator object for the magnetic_lenght: l=L(i)
	 * @param  name name of the kicker magnet
	 * @param  psname name of the power supply feeding the magnet
	 * @throw std::invalid_argument Interpolator object type not detected correctly
	 */
	Kicker(bool positive, Interpolator::Interpolator& direct_interpolator,Interpolator::Interpolator& inverse_interpolator,Interpolator::Interpolator& length_interpolator,std::string name="null",std::string psname="null");

	/**
	 * Destructor
	 */
	 ~Kicker( );

	/**
	 * @return the main magnetic field (integrated B)
	 * @param  excitation the magnet excitation ( current  )
	 * @throw std::range_error inherited from Field::Interpolator
	 */
	double excitationToField (const double excitation ) ;


	/**
	 * @return the magnetic length
	 * @param  excitation the excitation - current - driving the magnet.
	 * @throw std::range_error inherited from Field::Interpolator
	 */
	double excitationToMagneticLength(const double excitation ) ;



	/**
	 * @return the main magnet strength ( kick )
	 * @param  excitation excitation - current or voltage - driving the magnet
	 * @param  energyGeV the energy in GeV of the electron beam
	 * @throw std::range_error inherited from Field::Interpolator
	 */
	double excitationToStrength(const double excitation, const double energyGeV ) ;



	/**
	 * @return the required excitation ( current  )
	 * @param  field requested main magnetic field (integrated B)
	 * @throw std::range_error inherited from Field::Interpolator
	 */
	double fieldToExcitation(const double field ) ;



	/**
	 * @return the required excitation ( current )
	 * @param  strength requested magnet strength ( kick )
	 * @param  energyGeV the energy in GeV of the electron beam
	 * @throw std::range_error inherited from Field::Interpolator
	 */
	double strengthToExcitation(const double strength, const double energyGeV ) ;



	/**
	 * Get supported range of excitation values.
	 * Range is  [min,max).
	 * @param  min ( current  )
	 * @param  max ( current  )
	 */
	void getExcitationRange(double& min, double& max );




	/**
	 * Get supported range of main field values.
	 * Range is  [min,max).
	 * @param  min ( gradient )
	 * @param  max ( gradient )
	 */
	void getFieldRange(double& min, double& max );

	/**
	 * Get supported range of field strength values at a given energy.
	 * Range is  [min,max).
	 * @param  min ( k )
	 * @param  max ( k )
	 * @param  energy the energy in GeV of the electron beam
	 */

	void getStrengthRange(double& min, double& max, const double energy );
protected:

private:

};

} //close namespace Field;
#endif // FIELD_KICKER_H

