#ifndef FIELD_QUADRUPOLE_H
#define FIELD_QUADRUPOLE_H
#include "magnet.h"
#include <string>

/**
 * @class Quadrupole
 * Evaluates the main parameters of a quadrupole magnet: gradient,strength and magnetic length
 * or estimates the excitation required for a given gradient or strength.
 * Strength calculations are carried out for a given electron energy in GeV.
 * These parameters are ment to be used to calculate the transfer/tranport matrices of the magnet.
 * This class exploits various Interpolator objects to evaluate the different fields, which must be
 * constructed and inited correctly. A factory class will do this reading the data from the magnets database.
 *
 * Notice that by convention all the calibration curves hold POSITIVE values for POSITIVE excitations.
 * The effective sign of the field is controlled by a boolean parameter. The value of this parametr is also recorded
 * in the database and is handled by the factory class. focussing==true -> negative gradient -> flip sign of the field.
 * The name of the magnet and of the power supply are auxilary fields.
 */

/* $Author: claudio $
 *
 * $Revision: 1.13 $
 *
 * $Log: quadrupole.h,v $
 * Revision 1.13  2018-11-15 15:23:42  claudio
 * start release_8 developement
 *
 * Revision 1.12  2010/01/15 16:30:47  claudio
 * FocussingSignConvention: MAD (K>0) or Vicky (K<0) for quads and sexts.
 *
 * Revision 1.11  2009/03/31 07:57:18  claudio
 * minor doc corrections
 *
 * Revision 1.10  2009/01/07 08:42:57  claudio
 * adapted to new class names for interpolators
 *
 * Revision 1.9  2008/08/27 13:23:43  claudio
 * doxygen improvements
 *
 * Revision 1.8  2007/08/06 13:44:44  claudio
 * added functions to exactly calculate the inverse of the ineterpolator. based on the "regula falsi" root finding algorithm and
 * a bracleting strategy
 *
 * Revision 1.7  2007/07/24 13:09:52  claudio
 * cleanup of #ifdef guard against double inclusion
 * even across namespaces
 *
 * Revision 1.6  2007/07/09 09:13:53  claudio
 * Change som #includes for compatibility with canonical installation tree
 *
 * Revision 1.5  2007/05/18 12:36:41  claudio
 * refactored for sharing more internal variables and, later, code
 *
 * Revision 1.4  2007/03/30 12:55:51  claudio
 * Costrusice decentemente le classi base. Migliorata la documantazione. TODO: gestire meglio i range.
 * Mettere a fattore comune (rifattorizzare) il piu' possibile in magnet
 *
 * Revision 1.3  2007/03/28 14:55:37  claudio
 * refactor towards unification of common code
 *
 * Revision 1.2  2007/03/26 15:04:28  claudio
 * development step of end March
 *
 * Revision 1.1.1.1  2007/03/14 10:27:13  claudio
 * Hlf field module
 *
 */

namespace magnetfield
{

class Quadrupole : public Magnet
{
public:

	/**
	 * Constructor
	 * @param  focussing select focussing/defocussing field
	 * @param  direct_interpolator Interpolator object for the gradient: g=G(i)
	 * @param  inverse_interpolator Interpolator object for the current: i=I(g)
	 * @param  length_interpolator Interpolator object for the magnetic_lenght: l=L(i)
	 * @param  name name of the quadrupole magnet
	 * @param  psname name of the power supply feeding the magnet
	 * @throw std::invalid_argument Interpolator object type not detected correctly
	 */
	Quadrupole(bool focussing, Interpolator::Interpolator& direct_interpolator,Interpolator::Interpolator& inverse_interpolator,Interpolator::Interpolator& length_interpolator,std::string name="null",std::string psname="null");

	/**
	 * Destructor
	 */
	 ~Quadrupole( );

	/**
	 * @return the main magnetic field (gradient)
	 * @param  excitation the magnet excitation ( current  )
	 * @throw std::range_error inherited from Field::Interpolator
	 */
	double excitationToField (const double excitation );


	/**
	 * @return the magnetic length
	 * @param  excitation the excitation - current - driving the magnet.
	 * @throw std::range_error inherited from Field::Interpolator
	 */
	double excitationToMagneticLength(const double excitation );



	/**
	 * @return the main magnet strength (k)
	 * @param  excitation excitation - current or voltage - driving the magnet
	 * @param  energyGeV the energy in GeV of the electron beam
	 * @throw std::range_error inherited from Field::Interpolator
	 */
	double excitationToStrength(const double excitation, const double energyGeV );



	/**
	 * @return the required excitation ( current  )
	 * @param  field requested main magnetic field (gradient)
	 * @throw std::range_error inherited from Field::Interpolator
	 */
	double fieldToExcitation(const double field );



	/**
	 * @return the required excitation ( current )
	 * @param  strength requested magnet strength (k)
	 * @param  energyGeV the energy in GeV of the electron beam
	 * @throw std::range_error inherited from Field::Interpolator
	 */
	double strengthToExcitation(const double strength, const double energyGeV );



	/**
	 * Get supported range of excitation values.
	 * Range is  [min,max).
	 * @param  min ( current  )
	 * @param  max ( current  )
	 */
	void getExcitationRange(double& min, double& max );




	/**
	 * Get supported range of main field values.
	 * Range is  [min,max).
	 * @param  min ( gradient )
	 * @param  max ( gradient )
	 */
	void getFieldRange(double& min, double& max );

	/**
	 * Get supported range of field strength values at a given electron energy.
	 * Range is  [min,max).
	 * @param  min ( k )
	 * @param  max ( k )
	 * @param  energyGeV the energy in GeV of the electron beam
	 */

	void getStrengthRange(double& min, double& max, const double energyGeV );

	/**
	 * extend Magnet base class method
	 */
	void setFocSignConvention(focSignConvention);
protected:
	bool _actual_focussing;/**< true if magnet focussing in the horizontal plane means K <0 */

private:


};

} //close namespace Field;
#endif // FIELD_QUADRUPOLE_H

