#ifndef FIELD_SOLENOID_H
#define FIELD_SOLENOID_H
#include "magnet.h"
#include <string>

/**
 * @class Solenoid
 * Evaluates the main parameters of a solenoid magnet: filed and magnetic length
 * or estimates the excitation required for a given field.
 * Strength calculations are carried out for a given energy in GeV [electrons only!!!]
 * These parameters are ment to be used to calculate the transfer/tranport matrices of the magnet.
 * This class exploits various Interpolator objects to evaluate the different fields, which must be
 * constructed and inited correctly. A factory class will do this reading the data from the magnets database.
 *
 * Notice that by convention all the calibration curves hold POSITIVE values for POSITIVE excitations.
 * The effective sign of the field is controlled by a boolean parameter. The value of this parametr is also recorded
 * in the database and is handled by the factory class. focussing==true -> negative gradient -> flip sign of the field.
 * The name of the magnet and of the power supply are auxilary fields.
 */

/* $Author: claudio $
 *
 * $Revision: 1.2 $
 *
 * $Log: solenoid.h,v $
 * Revision 1.2  2018-11-15 15:23:42  claudio
 * start release_8 developement
 *
 * Revision 1.1  2009/03/24 09:09:38  claudio
 * first import, to be completed
 *
 */


namespace magnetfield
{

class Solenoid : public Magnet
{
public:

	/**
	 * Constructor
	 * @param  focussing select focussing/defocussing field
	 * @param  direct_interpolator Interpolator object for the gradient: g=G(i)
	 * @param  inverse_interpolator Interpolator object for the current: i=I(g)
	 * @param  length_interpolator Interpolator object for the magnetic_lenght: l=L(i)
	 * @param  name name of the solenoid magnet
	 * @param  psname name of the power supply feeding the magnet
	 * @throw std::invalid_argument Interpolator object type not detected correctly
	 */
	Solenoid(bool focussing, Interpolator::Interpolator& direct_interpolator,Interpolator::Interpolator& inverse_interpolator,Interpolator::Interpolator& length_interpolator,std::string name="null",std::string psname="null");

	/**
	 * Destructor
	 */
	 ~Solenoid( );

	/**
	 * @return the main magnetic field (gradient)
	 * @param  excitation the magnet excitation ( current  )
	 * @throw std::range_error inherited from Field::Interpolator
	 */
	double excitationToField (const double excitation );


	/**
	 * @return the magnetic length
	 * @param  excitation the excitation - current - driving the magnet.
	 * @throw std::range_error inherited from Field::Interpolator
	 */
	double excitationToMagneticLength(const double excitation );



	/**
	 * @return the main magnet strength (k)
	 * @param  excitation excitation - current or voltage - driving the magnet
	 * @param  energyGeV the energy in GeV of the electron beam
	 * @throw std::range_error inherited from Field::Interpolator
	 */
	double excitationToStrength(const double excitation, const double energyGeV );



	/**
	 * @return the required excitation ( current  )
	 * @param  field requested main magnetic field (gradient)
	 * @throw std::range_error inherited from Field::Interpolator
	 */
	double fieldToExcitation(const double field );



	/**
	 * @return the required excitation ( current )
	 * @param  strength requested magnet strength (k)
	 * @param  energyGeV the energy in GeV of the electron beam
	 * @throw std::range_error inherited from Field::Interpolator
	 */
	double strengthToExcitation(const double strength, const double energyGeV );



	/**
	 * Get supported range of excitation values.
	 * Range is  [min,max).
	 * @param  min ( current  )
	 * @param  max ( current  )
	 */
	void getExcitationRange(double& min, double& max );




	/**
	 * Get supported range of main field values.
	 * Range is  [min,max).
	 * @param  min ( gradient )
	 * @param  max ( gradient )
	 */
	void getFieldRange(double& min, double& max );

	/**
	 * Get supported range of field strength values at a given energy.
	 * Range is  [min,max).
	 * @param  min ( k )
	 * @param  max ( k )
	 * @param  energyGeV the energy in GeV of the electron beam
	 */

	void getStrengthRange(double& min, double& max, const double energyGeV );
protected:

private:


};

} //close namespace Field;
#endif // FIELD_SOLENOID_H

